#!/usr/bin/env python3
# -*- coding: UTF8 -*-

import datetime
from functools import partial
import itertools
import matplotlib.pyplot as plt
import os
import pickle
import sys
import time

import numpy as np
import scipy.spatial
import scipy.sparse

import scipy.sparse.csgraph as graph
from skimage.feature import peak_local_max
from sklearn.cluster import AgglomerativeClustering

import torch

import jax
import jax.numpy as jnp


def distance_one(x, y):
    return jnp.linalg.norm(x - y)


jax_cdist_0 = jax.vmap(distance_one, (None, 0))
jax_cdist = jax.vmap(jax_cdist_0, (0, None))
jax_cdist_0 = jax.jit(jax_cdist_0)
jax_cdist = jax.jit(jax_cdist)


class ArrayDataset(torch.utils.data.Dataset):
    def __init__(self, nparr):
        self.nparr = nparr

    def __len__(self):
        return len(self.nparr)

    def __getitem__(self, idx):
        return idx, self.nparr[idx]


def build_dataloader(dataset, num_workers, batch_size, shuffle=True):
    if not isinstance(dataset, torch.utils.data.Dataset) and not isinstance(dataset, torch.utils.data.DataLoader):
        dataset = ArrayDataset(dataset)
    if isinstance(dataset, torch.utils.data.Dataset):
        dataloader = torch.utils.data.DataLoader(dataset, batch_size=batch_size, shuffle=shuffle,
                                                 num_workers=num_workers, pin_memory=True)
    if isinstance(dataset, torch.utils.data.DataLoader):
        dataloader = dataset
    dataloader.collate_fn = jax_collate

    return dataloader


def jax_collate(items):
    """

    :param items: items is a list of tuples (id, vector)
    :return:
    """
    labels = [item[0] for item in items]
    np_data = [item[1] for item in items]
    np_data = np.stack(np_data)
    return labels, np_data


class SOM:
    """
    2-D Self-Organizing Map with Gaussian Neighbourhood function
    and linearly decreasing learning rate.

    Creates a grid of size m*n with centroids of dimensions dim randomly inititalised.
    precompute is longer to initialize but runs significantly faster
    use device to use a gpu
    Use device to use a gpu
    :param m: the height of the SOM
    :param n: the width of the SOM
    :param dim: the dimension of the input vectors
    :param alpha: The initial lr
    :param sigma: The initial width of the gaussian blur. If None, half of the grid is used
    :param niter: The number of epochs to run the SOM
    :param sched: Scheduler scheme. Possibilites are 'linear' for a linear decay, 'half' to half the lr 20 times and
                    'exp' for an exponential decay
    :param precompute: Speedup for initialization. Creates a little overhead for very small trainings
    :param periodic: Boolean to use a periodic topology
    :param metric: takes as input two arrays (n,p) and (m,p) and returns a distance matrix (n,m)
    :param centroids: Initial som map to use. No random initialization
     """

    def __init__(self,
                 m,
                 n,
                 dim,
                 alpha=None,
                 sigma=None,
                 n_epoch=2,
                 sched='linear',
                 device=jax.devices()[0],
                 precompute=True,
                 periodic=False,
                 metric=None,
                 centroids=None):

        # topology of the som
        super(SOM, self).__init__()
        self.device = device
        self.step = 0
        self.m = m
        self.n = n
        self.grid_size = m * n
        self.dim = dim
        self.periodic = periodic
        if metric is None:
            self.metric = jax_cdist
        else:
            self.metric = metric

        # optimization parameters
        self.sched = sched
        self.n_epoch = n_epoch
        if alpha is not None:
            self.alpha = float(alpha)
        else:
            self.alpha = alpha
        if sigma is None:
            self.sigma = np.sqrt(self.m * self.n) / 2.0
        else:
            self.sigma = float(sigma)

        if centroids is None:
            self.centroids = jax.device_put(np.random.randn(m * n, dim), device=device)
        else:
            self.centroids = centroids

        locs = np.stack([np.array([i, j]) for i in range(self.m) for j in range(self.n)])
        self.locations = jax.device_put(locs, device=device)
        self.precompute = precompute
        if self.precompute:
            # Fast computation is only right for the periodic topology
            if self.periodic:
                self.distance_mat = self.compute_all()
            else:
                self.distance_mat = jnp.stack([self.get_bmu_distance_squares(loc) for loc in self.locations])
            self.distance_mat = jax.device_put(self.distance_mat, device=device)
        self.umat = None

        # Clustering parameters
        self.cluster_att = None
        self.clusters_user = None

    def to_device(self, device):
        """
        Send the jax objects to device
        :param device:
        :return:
        """
        for k, v in vars(self).items():
            var = v
            if isinstance(var, jax.interpreters.xla._DeviceArray):
                var = jax.device_put(var, device=device)
                self.__dict__[k] = var
        return self

    # @jax.jit
    @partial(jax.jit, static_argnums=(0,))
    def get_bmu_distance_squares(self, bmu_loc):
        if self.periodic:
            offsets = jnp.array([[0, 0],
                                 [-self.m, -self.n],
                                 [self.m, self.n],
                                 [-self.m, 0],
                                 [self.m, 0],
                                 [0, -self.n],
                                 [0, self.n],
                                 [-self.m, self.n],
                                 [self.m, -self.n]])
            # offsets = jax.device_put(offsets)
            offset_locations = jnp.array([self.locations + offset for offset in offsets])
            batch_cdist = jax.vmap(jax_cdist_0, (None, 0))
            distances = batch_cdist(bmu_loc, offset_locations)
            min_dists = jnp.min(distances, axis=0)
            return min_dists
        else:
            distances = jax_cdist_0(bmu_loc, self.locations)
            return distances

    def compute_all(self):
        """
        Use the first line computation to get the other ones. Only works for periodic topology
        """

        # get first line of the grid this could be sped up with the same trick but indexing might get tricky
        # 'first_line' is thus the m first rows of the distance matrix and has shape m, grid_size
        import time
        first_line = [self.get_bmu_distance_squares(loc) for loc in self.locations[:self.m]]
        first_line = jnp.stack(first_line)
        # We use square euclidean distance.
        first_line = first_line ** 2
        other_lines = []
        for line in range(1, self.n):
            first_part = first_line[:, (self.grid_size - line * self.m):]
            second_part = first_line[:, :(self.grid_size - line * self.m)]
            block = jnp.concatenate((first_part, second_part), axis=1)
            other_lines.append(block)
        all_lines = jnp.concatenate((first_line, *other_lines), 0)
        return all_lines

    @staticmethod
    def find_batchsize(x):
        """
        Dimension needed is BS, 1(vector), dim
        """
        if len(x.shape) == 1:
            return x[None, None, :], 1
        batch_size = x.shape[0]
        return x[:, None, :], batch_size

    def scheduler(self, it, tot):
        if self.sched == 'linear':
            return 1.0 - it / tot
        # half the lr 20 times
        if self.sched == 'half':
            return 0.5 ** int(20 * it / tot)
        # decay from 1 to exp(-5)
        if self.sched == 'exp':
            return np.exp(-5 * it / tot)
        raise NotImplementedError('Wrong value of "sched"')

    def __call__(self, x, learning_rate_op):
        """
        timing info : now most of the time is in pdist ~1e-3s and the rest is 0.2e-3
        :param x: the minibatch
        :param learning_rate_op: the learning rate to apply to the batch
        :return:
        """
        bmu_loc, error, self.centroids = self.static_call(x, learning_rate_op, self.centroids)
        return bmu_loc, error

    @partial(jax.jit, static_argnums=(0,))
    def static_call(self, x, learning_rate_op, self_centroids):
        """
        timing info : now most of the time is in pdist ~1e-3s and the rest is 0.2e-3
        :param x: the minibatch
        :param learning_rate_op: the learning rate to apply to the batch
        :param self_centroids: to be compiled, the centroids cannot be fed through self since they will be muted.
        :return:
        """
        # Make an inference call
        # Compute distances from batch to centroids
        x, batch_size = self.find_batchsize(x)
        dists = self.metric(x, self_centroids)

        # x is of shape (30,1,50), centroids of shape (100, 50)

        # Find closest and retrieve the gaussian correlation matrix for each point in the batch
        # bmu_loc is BS, num points
        bmu_index = jnp.argmin(dists, -1)
        mindist = jnp.take_along_axis(dists, bmu_index[:, None], axis=-1)
        bmu_loc = self.locations[bmu_index].reshape(batch_size, 2)
        # Compute the update

        # Update LR
        # It is a matrix of shape (BS, centroids) and tell for each input how much it will affect each centroid
        alpha_op = self.alpha * learning_rate_op
        sigma_op = self.sigma * learning_rate_op
        if self.precompute:
            bmu_distance_squares = self.distance_mat[bmu_index].reshape(batch_size, self.grid_size)
        else:
            bmu_distance_squares = []
            for loc in bmu_loc:
                bmu_distance_squares.append(self.get_bmu_distance_squares(loc))
            bmu_distance_squares = jnp.stack(bmu_distance_squares)
        neighbourhood_func = jnp.exp(-(bmu_distance_squares / (2 * sigma_op ** 2 + 1e-5)))
        learning_rate_multiplier = alpha_op * neighbourhood_func

        # Take the difference of centroids with input and weight it with gaussian
        # x is (BS,1,dim)
        # self.weights is (grid_size,dim)
        # delta is (BS, grid_size, dim)
        def flat_form(one_x, alpha, weight):
            small_d = one_x - weight
            mutliplied = small_d * alpha[:, None]
            return mutliplied

        vdelta = jax.vmap(flat_form, in_axes=(0, 0, None))
        delta = vdelta(x, learning_rate_multiplier, self_centroids)
        # Perform the update by taking the mean
        delta = jnp.mean(delta, axis=0)
        new_weights = self_centroids + delta
        return bmu_loc, jnp.mean(mindist), new_weights

    # @partial(jax.jit, static_argnums=(0,))
    def inference_call_one(self, x):
        """
        timing info : now most of the time is in pdist ~1e-3s and the rest is 0.2e-3
        :param x:
        :param it:
        :return:
        """
        # Compute distances from batch to centroids
        # Dimension needed is BS, 1(vector), dim
        x, batch_size = self.find_batchsize(x)
        dists = self.metric(x, self.centroids)

        # Find closest and retrieve the gaussian correlation matrix for each point in the batch
        bmu_index = jnp.argmin(dists, -1)
        mindist = jnp.take_along_axis(dists, bmu_index[:, None], axis=-1)
        return bmu_index, mindist

    # @partial(jax.jit, static_argnums=(0, 2))
    def inference_call_two(self, x, n_bmu=2):
        """
        timing info : now most of the time is in pdist ~1e-3s and the rest is 0.2e-3
        :param x:
        :param it:
        :return:
        """
        # Compute distances from batch to centroids
        x, batch_size = self.find_batchsize(x)
        dists = self.metric(x, self.centroids)
        idx = np.argsort(dists, -1)
        selected = idx[:, :1]
        mindists = jnp.take_along_axis(dists, selected, axis=-1)
        return selected, mindists

    def inference_call(self, x, n_bmu=1):
        """
        timing info : now most of the time is in pdist ~1e-3s and the rest is 0.2e-3
        :param x:
        :param it:
        :return:
        """
        # Find closest and retrieve the gaussian correlation matrix for each point in the batch
        if n_bmu == 1:
            return self.inference_call_one(x)
        else:
            # In that case, return the bmu indices.
            return self.inference_call_two(x, n_bmu)

    def loc_from_idx(self, idx):
        loc = self.locations[idx]
        loc = jnp.reshape(loc, (-1, 2))
        return loc

    def fit(self,
            dataset=None,
            batch_size=20,
            n_epoch=None,
            print_each=100,
            do_compute_all_dists=True,
            unfold=True,
            normalize_umat=True,
            sigma=None,
            alpha=None,
            logfile='som.log',
            num_workers=os.cpu_count()):
        """
        samples: torch tensor with all the data. If given dataloader must not be given
        dataset: torch data loader object. If given samples must not be given
        """
        if logfile is not None:
            logfile = open(logfile, 'w', buffering=1)
            logfile.write('#epoch #iter #alpha #sigma #error #runtime\n')
        dataloader = build_dataloader(dataset, num_workers, batch_size=batch_size)
        nbatch = len(dataloader)

        if self.alpha is None:
            self.alpha = float((self.m * self.n) / nbatch)
            print('alpha:', self.alpha)
        if sigma is not None:
            # reset the sigma
            self.sigma = sigma
        if alpha is not None:
            # reset the alpha
            self.alpha = alpha
        if n_epoch is None:
            n_epoch = self.n_epoch
        npts = len(dataloader.dataset)
        total_steps = npts * n_epoch
        start = time.perf_counter()
        learning_error = list()
        for epoch in range(n_epoch):
            for i, (label, batch) in enumerate(dataloader):
                lr_step = self.scheduler(self.step, total_steps)
                batch = jax.device_put(batch, self.device)
                bmu_loc, error, new_centroids = self.static_call(batch,
                                                                 learning_rate_op=lr_step,
                                                                 self_centroids=self.centroids)
                self.centroids = new_centroids
                learning_error.append(error)
                if not i % print_each:
                    runtime = time.perf_counter() - start
                    eta = total_steps * runtime / (self.step + batch_size) - runtime
                    print(f'{epoch + 1}/{n_epoch}: {self.step}/{total_steps} '
                          f'| alpha: {self.alpha * lr_step:4f} | sigma: {self.sigma * lr_step:4f} '
                          f'| error: {error:4f} | time {runtime:4f}'
                          f'| eta: {str(datetime.timedelta(seconds=eta))}',
                          flush=True)
                    if logfile is not None:
                        logfile.write(
                            f'{epoch} {self.step} {self.alpha * lr_step} {self.sigma * lr_step} {error} {runtime}\n')
                self.step += batch_size
        self.compute_umat(unfold=unfold, normalize=normalize_umat)
        if do_compute_all_dists:
            self.compute_all_dists()
        if logfile is not None:
            logfile.close()
        return learning_error

    def predict(self, dataset, batch_size=100, print_each=100,
                return_density=False, return_errors=False, num_workers=os.cpu_count()):
        """
        Batch the prediction to avoid memory overloading
        """
        dataloader = build_dataloader(dataset, num_workers=num_workers, batch_size=batch_size, shuffle=False)
        bmus = list()
        errors = list()
        if return_density:
            density = np.zeros((self.m, self.n))
        if return_errors:
            bmu_indices = []
        labels = []
        start = time.perf_counter()
        for i, (label, batch) in enumerate(dataloader):
            batch = jax.device_put(batch, self.device)

            # If we want the topographic error, we need to compute more neighbors, and keep the indices
            bmu_idx, error = self.inference_call(batch, n_bmu=2 if return_errors else 1)
            if return_errors:
                bmu_indices.append(bmu_idx)

            # Then we can keep the first only and compute its bmu affectation
            first_idx = bmu_idx[:, 0] if return_errors else bmu_idx
            bmu_loc = self.loc_from_idx(first_idx)
            labels.extend(label)
            bmus.append(bmu_loc)
            if return_density:
                density[tuple(np.asarray(bmu_loc).T)] += 1.
            if error.ndim == 0:
                error = error[None, ...]
            errors.append(error)
        bmus = jnp.concatenate(bmus)
        bmus = np.asarray(bmus)
        errors = jnp.concatenate(errors)
        errors = np.asarray(errors).flatten()

        default_return = [bmus, errors, labels]
        if return_density:
            density /= density.sum()
            default_return.append(density)
        # Optionnally compute errors
        if return_errors:
            quantization_error = np.mean(errors[:, :, 0])
            topo_dists = np.array([self.distance_mat[int(first), int(second)] for first, second in bmu_indices])
            topo_error = np.sum(topo_dists > 1) / len(topo_dists)
            print(f'On these samples, the quantization error is {quantization_error:1f} '
                  f'and the topological error rate is {topo_error:1f}')
            default_return.append(quantization_error)
            default_return.append(topo_error)
        return tuple(default_return)

    def plot_component_plane(self, plane, savefig=None, show=True):
        """
        Get component plane plot : The value of each dimension for each centroid
        """
        smap = self.centroids.cpu().numpy().reshape((self.m, self.n, -1))
        assert 0 <= plane < self.dim
        component_plane = smap[:, :, plane]

        plt.matshow(component_plane)
        plt.colorbar()
        if savefig is not None:
            plt.savefig(savefig)
        if show:
            plt.show()
        plt.clf()

    @staticmethod
    def _unfold(umat, adj):
        """
        - umat: U-matrix
        - adj: Adjacency matrix of the corresponding U-matrix
        """
        n1, n2 = umat.shape
        mstree = graph.minimum_spanning_tree(adj)
        start = umat.argmin()
        sdist, pred = graph.shortest_path(mstree, indices=start, directed=False, return_predecessors=True)
        floodpath = np.asarray(np.unravel_index(sdist.argsort(), (n1, n2))).T
        parents_floodpath = np.asarray(np.unravel_index(pred[sdist.argsort()][1:], (n1, n2))).T
        flow = np.sign(floodpath[1:] - parents_floodpath)
        i, j = floodpath[0]
        u, v = 0, 0
        mapping = {(i, j): (u, v)}
        u_min, v_min = 0, 0

        for step, vec in enumerate(flow):
            i, j = floodpath[step + 1]
            u, v = mapping[tuple(parents_floodpath[step])] + vec
            u_min = u if u < u_min else u_min
            v_min = v if v < v_min else v_min
            mapping[(i, j)] = (u, v)
        path = [(u - u_min, v - v_min) for u, v in mapping.values()]
        floodpath = [(i, j) for i, j in floodpath]
        mapping = dict(zip(floodpath, path))
        uumat = np.ones(np.asarray(path).max(axis=0) + 1) * np.inf
        # uumat = np.ones(np.asarray(path).max(axis=0) + 1)
        for k, v in mapping.items():
            uumat[v[0], v[1]] = umat[k[0], k[1]]
        return uumat, mapping

    def _get_umat(self, smap, shape=None, rmsd=False, return_adjacency=False, periodic=True):
        """
        Compute the U-matrix based on a map of centroids and their connectivity.
        """

        def neighbor_dim2_toric(p, s):
            """
            Efficient toric neighborhood function for 2D SOM.
            """
            x, y = p
            X, Y = s
            xm = (x - 1) % X
            ym = (y - 1) % Y
            xp = (x + 1) % X
            yp = (y + 1) % Y
            return [(xm, ym), (xm, y), (xm, yp), (x, ym), (x, yp), (xp, ym), (xp, y), (xp, yp)]

        def neighbor_dim2_grid(p, s):
            """
            Efficient grid neighborhood function for 2D SOM.
            """
            x, y = p
            X, Y = s
            xm = max((x - 1), 0)
            ym = max((y - 1), 0)
            xp = min((x + 1), X - 1)
            yp = min((y + 1), Y - 1)
            shortlist = {(xm, ym), (xm, y), (xm, yp), (x, ym), (x, yp), (xp, ym), (xp, y), (xp, yp)}
            shortlist.discard((x, y))
            return list(shortlist)

        # By default a map is a square
        if shape is None:
            if smap.ndim == 2:
                n = int(np.sqrt(smap.shape[0]))
                smap = smap.reshape((n, n, -1))
            shape = list(smap.shape)[:-1]
        umatrix = np.zeros(shape)
        adjmat = {'data': [], 'row': [], 'col': []}

        start = time.perf_counter()
        for i, point in enumerate(itertools.product(*[range(s) for s in shape])):
            # if not i % 100:
            #     print(f'{i} | time: {str(datetime.timedelta(seconds=time.perf_counter() - start))} ')
            neuron = smap[point][None, :]
            if periodic:
                neighbors = tuple(np.asarray(neighbor_dim2_toric(point, shape), dtype='int').T)
            else:
                neighbors = tuple(np.asarray(neighbor_dim2_grid(point, shape), dtype='int').T)
            smap_neighbor = smap[neighbors]
            cdists = np.asarray(self.metric(smap_neighbor, neuron))
            umatrix[point] = cdists.mean()

            adjmat['row'].extend([np.ravel_multi_index(point, shape), ] * len(neighbors[0]))
            adjmat['col'].extend(np.ravel_multi_index(neighbors, shape))
            adjmat['data'].extend(cdists[:, 0])
        if rmsd:
            natoms = smap.shape[-1] / 3.
            umatrix /= natoms
            umatrix = np.sqrt(umatrix)

        if return_adjacency:
            adjmat = scipy.sparse.coo_matrix((adjmat['data'], (adjmat['row'], adjmat['col'])),
                                             shape=(np.prod(shape), np.prod(shape)))
            return umatrix, adjmat
        else:
            return umatrix

    def compute_umat(self, unfold=True, normalize=True):
        smap = np.asarray(self.centroids).reshape((self.m, self.n, -1))
        umat, adj = self._get_umat(smap=smap, shape=(self.m, self.n),
                                   return_adjacency=True, periodic=self.periodic)
        if normalize:
            # Renormalize
            umat = (umat - np.min(umat)) / (np.max(umat) - np.min(umat))
        self.umat = umat
        self.adj = adj
        if self.periodic and unfold:
            uumat, mapping = self._unfold(umat, adj)
            self.uumat = uumat
            self.mapping = mapping
            self.reversed_mapping = {v: k for k, v in self.mapping.items()}
            umat = uumat
        else:
            self.mapping = {(i, j): (i, j) for (i, j) in itertools.product(range(self.m), range(self.n))}
            self.reversed_mapping = {v: k for k, v in self.mapping.items()}
            self.uumat = umat

    def compute_all_dists(self):
        # GRAPH-BASED
        # mstree = graph.minimum_spanning_tree(self.adj)
        adj = self.adj.tocsr()
        self.all_to_all_dist = graph.shortest_path(adj, directed=False)
        # self.all_to_all_dist = graph.shortest_path(mstree, directed=False)

    def cluster(self, min_distance=2):
        """
        Perform clustering based on the umatrix.

        We have tried several methods that can be put into two main categories :
        Graph-based : using either a minimum spanning tree of the full connectivity
        Image-based : using the U-matrix and segmentation techniques. In the periodic case,
                        an unfolding of the umatrix was necessary to do so.
        """

        local_min = peak_local_max(-self.umat, min_distance=min_distance)
        n_local_min = local_min.shape[0]
        clusterer = AgglomerativeClustering(affinity='precomputed', linkage='average', n_clusters=n_local_min)
        try:
            labels = clusterer.fit_predict(self.all_to_all_dist)
        except ValueError as e:
            print(f'WARNING : The following error was catched : "{e}"\n'
                  f'The clusterer yields zero clusters on the data.'
                  ' You should train it more or gather more data')
            labels = np.zeros(self.m * self.n)
        labels = labels.reshape((self.m, self.n))

        # IMAGE-BASED
        # from skimage.morphology import disk
        # from skimage.filters import rank
        # from skimage.util import img_as_ubyte
        # from skimage.segmentation import watershed
        # # denoise image
        # umat = img_as_ubyte(umat)
        # denoised = rank.median(umat, disk(2))
        # # find continuous region (low gradient -
        # # where less than 10 for this image) --> markers
        # # disk(5) is used here to get a more smooth image
        # markers = rank.gradient(denoised, disk(5)) < 10
        # markers = ndi.label(markers)[0]
        # # local gradient (disk(2) is used to keep edges thin)
        # gradient = rank.gradient(denoised, disk(2))
        # labels = watershed(gradient, markers)
        # if self.periodic:
        #     square_label = np.zeros((self.m, self.n))
        #     for (u, v), (i, j) in self.reversed_mapping.items():
        #         square_label[i, j] = labels[u, v]
        #     return square_label
        self.cluster_att = labels.flatten()
        return labels

    def manual_cluster(self, autocluster=False):
        # from .somgui import Wheel, Click

        fig, ax = plt.subplots(figsize=(8, 10))
        cax = ax.matshow(self.umat)
        fig.colorbar(cax)
        plt.show()

        click = Click(ax=ax)
        fig.canvas.mpl_connect('button_press_event', click)
        wheel = Wheel(self, click, ax=ax)

        if autocluster:
            self.clusters_user = self.cluster() + 1
        if self.clusters_user is not None:
            wheel.clusters = self.clusters_user
            wheel.expand_clusters()
            wheel.plot_clusters()
        fig.canvas.mpl_connect('scroll_event', wheel)
        fig.canvas.mpl_connect('button_press_event', wheel)
        ax.format_coord = wheel.format_coord
        plt.show()
        self.cluster_att = wheel.expanded_clusters.flatten()
        self.clusters_user = wheel.clusters

    def predict_cluster(self, samples=None, batch_size=100, user=False):
        """
        we have a mapping from each unit to its cluster in the flattened form in self.cluster_att
        Then we need to turn the bmu attributions into the index in this list and return the cluster attributions
        """
        if self.cluster_att is None:
            cluster_att = self.cluster()
            self.cluster_att = cluster_att.flatten()
        if samples is None:
            try:
                self.bmus
            except:
                print('No existing BMUs in the SOM object, one needs data points to predict clusters on')
        else:
            bmus, error, labels = self.predict(samples, batch_size=batch_size)
            self.bmus = bmus
            self.error = error
        flat_bmus = (self.bmus[:, 0] * self.n + self.bmus[:, 1]).astype(np.int32)
        codebook = []
        inds = np.arange(len(self.bmus))
        for i in range(self.m * self.n):
            sel = (flat_bmus == i)
            if sel.sum() > 0:
                ind = inds[sel][self.error[sel].argmin()]
                codebook.append(ind)
            else:
                codebook.append(-1)
        self.codebook = codebook
        if not user:
            return self.cluster_att[flat_bmus], error
        else:
            return self.clusters_user.flatten()[flat_bmus], error

    def save_pickle(self, outname):
        del self.device
        del self.metric
        pickle.dump(self, open(outname, 'wb'))

    @staticmethod
    def load_pickle(inname, metric=None, device=jax.devices()[0]):
        loaded_som = pickle.load(open(inname, 'rb'))
        if metric is None:
            print('WARNING : loading a JAX SOM object without its metric results in the cdist metric being used '
                  'because jitted functions can not be pickled. If you want to use another metric, please '
                  'specify it in the SOM.load_pickle function')
            metric = jax_cdist
        loaded_som.device = device
        loaded_som.metric = metric

        # We need to manually choose which arrays are to be put, otherwise every numpy item ends up a jax object.
        loaded_som.centroids = jax.device_put(loaded_som.centroids, device=device)
        loaded_som.locations = jax.device_put(loaded_som.locations, device=device)
        loaded_som.distance_mat = jax.device_put(loaded_som.distance_mat, device=device)
        return loaded_som


def time_som(som, X):
    som.alpha = 0.5
    import time

    X = jax.device_put(X)
    a = time.perf_counter()
    for _ in range(1000):
        som(X[:30], 1)
    jax.block_until_ready(som.centroids)
    print('total time : ', time.perf_counter() - a)

    som.alpha = 0.5
    import time

    a = time.perf_counter()
    # First call is compilation
    _, _, new_centroids = som.static_call(X[:30], 1, som.centroids)
    som.centroids = new_centroids
    jax.block_until_ready(som.centroids)
    print('compilation time : ', time.perf_counter() - a)

    # Second call is runtime
    a = time.perf_counter()
    for _ in range(1000):
        _, _, new_centroids = som.static_call(X[:30], 1, som.centroids)
        som.centroids = new_centroids
    jax.block_until_ready(som.centroids)
    print('total time : ', time.perf_counter() - a)
    sys.exit()


if __name__ == '__main__':
    pass
    # Prepare data
    np.random.seed(0)
    X = np.random.rand(1000, 500)

    # Create SOM
    n = 10
    somsize = n ** 2
    nsamples = X.shape[0]
    dim = X.shape[1]
    niter = 5
    batch_size = 30
    nsteps = int(nsamples / batch_size)
    som = SOM(n, n, dim, n_epoch=niter, precompute=True, periodic=True)
    som.to_device(jax.devices()[0])

    # Fit it and get results
    learning_error = som.fit(X, batch_size=batch_size)
    bmus, inference_error, labels = som.predict(X, batch_size=batch_size)
    predicted_clusts, errors = som.predict_cluster(X[45:56])
    print('some cluster for some random points are : ', predicted_clusts)



#!/usr/bin/env python3
# -*- coding: UTF8 -*-

#############################################################################
# Authors: Vincent Mallet, Guillaume Bouvier	      	                    #
# Copyright (c) 2021 Institut Pasteur                                       #
#                 				                                            #
#                                                                           #
#  Redistribution and use in source and binary forms, with or without       #
#  modification, are permitted provided that the following conditions       #
#  are met:                                                                 #
#                                                                           #
#  1. Redistributions of source code must retain the above copyright        #
#  notice, this list of conditions and the following disclaimer.            #
#  2. Redistributions in binary form must reproduce the above copyright     #
#  notice, this list of conditions and the following disclaimer in the      #
#  documentation and/or other materials provided with the distribution.     #
#  3. Neither the name of the copyright holder nor the names of its         #
#  contributors may be used to endorse or promote products derived from     #
#  this software without specific prior written permission.                 #
#                                                                           #
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS      #
#  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT        #
#  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR    #
#  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT     #
#  HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,   #
#  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT         #
#  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,    #
#  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY    #
#  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT      #
#  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE    #
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.     #
#                                                                           #
#  This program is free software: you can redistribute it and/or modify     #
#                                                                           #
#############################################################################

import os
import sys
import datetime
import functools
import itertools
import matplotlib.pyplot as plt
import pickle
import time

import numpy as np
import scipy.spatial
import scipy.sparse
import scipy.sparse.csgraph as graph
from skimage.feature import peak_local_max
from sklearn.cluster import AgglomerativeClustering
from sklearn.manifold import MDS

import torch
import torch.nn as nn


class ArrayDataset(torch.utils.data.Dataset):
    def __init__(self, nparr):
        self.nparr = nparr

    def __len__(self):
        return len(self.nparr)

    def __getitem__(self, idx):
        return idx, self.nparr[idx]


def build_dataloader(dataset, num_workers, batch_size, shuffle=True):
    if not isinstance(dataset, torch.utils.data.Dataset) and not isinstance(dataset, torch.utils.data.DataLoader):
        dataset = ArrayDataset(dataset)
    if isinstance(dataset, torch.utils.data.Dataset):
        dataloader = torch.utils.data.DataLoader(dataset,
                                                 batch_size=batch_size,
                                                 shuffle=shuffle,
                                                 pin_memory=True,
                                                 num_workers=num_workers)
    if isinstance(dataset, torch.utils.data.DataLoader):
        dataloader = dataset
    return dataloader


def check_symmetric(a):
    """
    Check if array a is symmetric
    """
    relative_error = np.sqrt(np.mean((a - a.T) ** 2)) / np.sqrt((a ** 2).mean())
    return relative_error


def symmetrize(a):
    """
    symmetrize array a. Return the symmetrized array and the relative error
    """
    a_sym = (a + a.T) / 2.
    relative_error = np.sqrt(np.mean((a - a_sym) ** 2)) / np.sqrt((a ** 2).mean())
    return a_sym, relative_error


class SOM(nn.Module):
    """
    2-D Self-Organizing Map with Gaussian Neighbourhood function
    and linearly decreasing learning rate.

    Creates a grid of size m*n with centroids of dimensions dim randomly inititalised.
    precompute is longer to initialize but runs significantly faster
    use device to use a gpu
    Use device to use a gpu
    :param m: the height of the SOM
    :param n: the width of the SOM
    :param dim: the dimension of the input vectors
    :param alpha: The initial lr
    :param sigma: The initial width of the gaussian blur. If None, half of the grid is used
    :param niter: The number of epochs to run the SOM
    :param sched: Scheduler scheme. Possibilites are 'linear' for a linear decay, 'half' to half the lr 20 times and
                    'exp' for an exponential decay
    :param device: the torch device to create the SOM onto. This can be modified using the to() method
    :param precompute: Speedup for initialization. Creates a little overhead for very small trainings
    :param periodic: Boolean to use a periodic topology
    :param metric: takes as input two torch arrays (n,p) and (m,p) and returns a distance matrix (n,m)
    :param p_norm: p value for the p-norm distance to calculate between each vector pair for torch.cdist
    :param centroids: Initial som map to use. No random initialization
     """

    def __init__(self,
                 m,
                 n,
                 dim,
                 alpha=None,
                 sigma=None,
                 n_epoch=2,
                 sched='linear',
                 device='cpu',
                 precompute=True,
                 periodic=False,
                 metric=None,
                 p_norm=2,
                 centroids=None):

        # topology of the som
        super(SOM, self).__init__()
        self.device = device
        self.step = 0
        self.m = m
        self.n = n
        self.grid_size = m * n
        self.dim = dim
        self.periodic = periodic
        self.p_norm = p_norm
        if metric is None:
            self.metric = functools.partial(torch.cdist, p=self.p_norm)
        else:
            self.metric = metric
        self.pairwise_dist = None
        self.bmus = None
        self.mds = None
        # optimization parameters
        self.sched = sched
        self.n_epoch = n_epoch
        if alpha is not None:
            self.alpha = float(alpha)
        else:
            self.alpha = alpha
        if sigma is None:
            self.sigma = np.sqrt(self.m * self.n) / 2.0
        else:
            self.sigma = float(sigma)

        if centroids is None:
            np_init = np.random.randn(m * n, dim)
            self.centroids = torch.from_numpy(np_init).float().to(device=device)
            # self.centroids = torch.randn(m * n, dim, device=device, dtype=torch.float)
        else:
            self.centroids = centroids
        # self.centroids = torch.randn(m * n, dim, device=device, dtype=torch.double)

        locs = [np.array([i, j]) for i in range(self.m) for j in range(self.n)]
        self.locations = torch.LongTensor(np.array(locs)).to(device)
        self.maprange = torch.stack([torch.tensor((self.m, self.n)) for i in range(self.m * self.n)]).float().to(device)

        self.offset1 = torch.tensor([-self.m, -self.n], device=device)
        self.offset2 = torch.tensor([self.m, self.n], device=device)
        self.offset3 = torch.tensor([-self.m, 0], device=device)
        self.offset4 = torch.tensor([self.m, 0], device=device)
        self.offset5 = torch.tensor([0, -self.n], device=device)
        self.offset6 = torch.tensor([0, self.n], device=device)
        self.offset7 = torch.tensor([-self.m, self.n], device=device)
        self.offset8 = torch.tensor([self.m, -self.n], device=device)

        self.precompute = precompute
        if self.precompute:
            # Fast computation is only right for the periodic topology
            if self.periodic:
                self.distance_mat = self.compute_all()
            else:
                self.distance_mat = torch.stack([self.get_bmu_distance_squares(loc) for loc in self.locations])
        self.umat = None

        # Clustering parameters
        self.cluster_att = None
        self.clusters_user = None

    def to_device(self, device):
        self.device = device
        for k, v in vars(self).items():
            var = v
            try:
                var = var.to(device)
                self.__dict__[k] = var
                # print(f'{k} -> {device}')
            except AttributeError:
                pass
                # print(f'{k} ... {device}')
        return self

    def get_bmu_distance_squares(self, bmu_loc):
        bmu_loc = bmu_loc.unsqueeze(0).expand_as(self.locations).float()
        if self.periodic:
            d1 = torch.sum(torch.pow(self.locations.float() - bmu_loc, 2), 1)
            d2 = torch.sum(torch.pow(self.locations.float() + self.offset1 - bmu_loc, 2), 1)
            d3 = torch.sum(torch.pow(self.locations.float() + self.offset2 - bmu_loc, 2), 1)
            d4 = torch.sum(torch.pow(self.locations.float() + self.offset3 - bmu_loc, 2), 1)
            d5 = torch.sum(torch.pow(self.locations.float() + self.offset4 - bmu_loc, 2), 1)
            d6 = torch.sum(torch.pow(self.locations.float() + self.offset5 - bmu_loc, 2), 1)
            d7 = torch.sum(torch.pow(self.locations.float() + self.offset6 - bmu_loc, 2), 1)
            d8 = torch.sum(torch.pow(self.locations.float() + self.offset7 - bmu_loc, 2), 1)
            d9 = torch.sum(torch.pow(self.locations.float() + self.offset8 - bmu_loc, 2), 1)
            bmu_distance_squares, _ = torch.min(torch.stack([d1, d2, d3, d4, d5, d6, d7, d8, d9]), 0)
        else:
            bmu_distance_squares = torch.sum(torch.pow(self.locations.float() - bmu_loc, 2), 1)
        return bmu_distance_squares

    def compute_all(self):
        """use the first line computation to get the other ones. Only works for periodic topology"""

        # get first line of the grid this could be sped up with the same trick but indexing might get tricky
        # 'first_line' is thus the m first rows of the distance matrix and has shape m, grid_size
        first_line = [self.get_bmu_distance_squares(loc) for loc in self.locations[:self.m]]
        first_line = torch.cat(first_line)
        first_line = first_line.reshape((self.m, self.grid_size))
        other_lines = []
        for line in range(1, self.n):
            a = first_line[:, (self.grid_size - line * self.m):]
            b = first_line[:, :(self.grid_size - line * self.m)]
            block = torch.cat((a, b), dim=1)
            other_lines.append(block)
        all_lines = torch.cat((first_line, *other_lines), 0)
        return all_lines

    @staticmethod
    def find_batchsize(x):
        """
        Dimension needed is BS, 1(vector), dim
        """
        if len(x.size()) == 1:
            x.unsqueeze_(0)
        batch_size = x.size()[0]
        return x[:, None, :], batch_size

    def scheduler(self, it, tot):
        if self.sched == 'linear':
            return 1.0 - it / tot
        # half the lr 20 times
        if self.sched == 'half':
            return 0.5 ** int(20 * it / tot)
        # decay from 1 to exp(-5)
        if self.sched == 'exp':
            return np.exp(-5 * it / tot)
        raise NotImplementedError('Wrong value of "sched"')

    def __call__(self, x, learning_rate_op):
        """
        timing info : now most of the time is in pdist ~1e-3s and the rest is 0.2e-3
        :param x: the minibatch
        :param learning_rate_op: the learning rate to apply to the batch
        :return:
        """
        # Make an inference call
        # Compute distances from batch to centroids
        x, batch_size = self.find_batchsize(x)
        dists = self.metric(x, self.centroids)
        # Find closest and retrieve the gaussian correlation matrix for each point in the batch
        # bmu_loc is BS, num points
        mindist, bmu_index = torch.min(dists, -1)
        bmu_loc = self.locations[bmu_index].reshape(batch_size, 2)

        # Compute the update

        # Update LR
        # It is a matrix of shape (BS, centroids) and tell for each input how much it will affect each centroid
        alpha_op = self.alpha * learning_rate_op
        sigma_op = self.sigma * learning_rate_op
        self.alpha_op = alpha_op
        self.sigma_op = sigma_op
        if self.precompute:
            bmu_distance_squares = self.distance_mat[bmu_index].reshape(batch_size, self.grid_size)
        else:
            bmu_distance_squares = []
            for loc in bmu_loc:
                bmu_distance_squares.append(self.get_bmu_distance_squares(loc))
            bmu_distance_squares = torch.stack(bmu_distance_squares)
        neighbourhood_func = torch.exp(torch.neg(torch.div(bmu_distance_squares, 2 * sigma_op ** 2 + 1e-5)))
        learning_rate_multiplier = alpha_op * neighbourhood_func

        # Take the difference of centroids with input and weight it with gaussian
        # x is (BS,1,dim)
        # self.weights is (grid_size,dim)
        # delta is (BS, grid_size, dim)
        expanded_x = x.expand(-1, self.grid_size, -1)
        expanded_weights = self.centroids.unsqueeze(0).expand((batch_size, -1, -1))
        delta = expanded_x - expanded_weights
        delta = torch.mul(learning_rate_multiplier.reshape(*learning_rate_multiplier.size(), 1).expand_as(delta), delta)

        # import functorch
        #
        # def flat_form(one_x, alpha, weight):
        #     small_d = one_x - weight
        #     mutliplied = small_d * alpha[:, None]
        #     return mutliplied
        # vdelta = functorch.vmap(flat_form, in_dims=(0, 0, None))
        # delta_2 = vdelta(x, learning_rate_multiplier, self.centroids)
        # diff = delta - delta_2
        # print(diff.mean())

        # Perform the update by taking the mean
        delta = torch.mean(delta, dim=0)
        new_weights = torch.add(self.centroids, delta)
        self.centroids = new_weights
        return bmu_loc, torch.mean(mindist)

    def inference_call(self, x, n_bmu=1):
        """
        timing info : now most of the time is in pdist ~1e-3s and the rest is 0.2e-3
        :param x:
        :param it:
        :return:
        """
        # Compute distances from batch to centroids
        # Dimension needed is BS, 1(vector), dim
        x, batch_size = self.find_batchsize(x)
        dists = self.metric(x, self.centroids)

        # Find closest and retrieve the gaussian correlation matrix for each point in the batch
        # bmu_loc is BS, num points
        if n_bmu == 1:
            mindist, bmu_index = torch.min(dists, -1)
            return bmu_index, mindist
        else:
            # In that case, return the bmu indices.
            idx = torch.argsort(dists, -1)
            selected = idx[:, :n_bmu]
            mindists = torch.take(dists, selected)
            return selected.squeeze(), mindists

    def fit(self,
            dataset=None,
            batch_size=20,
            n_epoch=None,
            print_each=100,
            do_compute_all_dists=True,
            unfold=True,
            normalize_umat=True,
            sigma=None,
            alpha=None,
            logfile='som.log',
            num_workers=os.cpu_count()):
        """
        samples: torch tensor with all the data. If given dataloader must not be given
        dataset: torch data loader object. If given samples must not be given
        """
        if logfile is not None:
            logfile = open(logfile, 'w', buffering=1)
            logfile.write('#epoch #iter #alpha #sigma #error #runtime\n')
        dataloader = build_dataloader(dataset, num_workers, batch_size=batch_size)
        nbatch = len(dataloader)

        if self.alpha is None:
            self.alpha = float((self.m * self.n) / nbatch)
            print('alpha:', self.alpha)
        if sigma is not None:
            # reset the sigma
            self.sigma = sigma
        if alpha is not None:
            # reset the alpha
            self.alpha = alpha
        if n_epoch is None:
            n_epoch = self.n_epoch
        npts = len(dataloader.dataset)
        total_steps = npts * n_epoch
        start = time.perf_counter()
        learning_error = list()
        for epoch in range(n_epoch):
            for i, (label, batch) in enumerate(dataloader):
                lr_step = self.scheduler(self.step, total_steps)
                batch = batch.to(self.device, non_blocking=True)
                batch = batch.float()
                bmu_loc, error = self.__call__(batch, learning_rate_op=lr_step)
                learning_error.append(error)
                if not i % print_each:
                    runtime = time.perf_counter() - start
                    eta = total_steps * runtime / (self.step + batch_size) - runtime
                    print(
                        f'{epoch + 1}/{n_epoch}: {self.step}/{total_steps} '
                        f'| alpha: {self.alpha_op:4f} | sigma: {self.sigma_op:4f} '
                        f'| error: {error:4f} | time: {str(datetime.timedelta(seconds=runtime))} '
                        f'| eta: {str(datetime.timedelta(seconds=eta))}',
                        flush=True)
                    if logfile is not None:
                        logfile.write(f'{epoch} {self.step} {self.alpha_op} {self.sigma_op} {error} {runtime}\n')
                self.step += batch_size
                # if self.step > 10 * batch_size:
                #     sys.exit()
        self.compute_umat(unfold=unfold, normalize=normalize_umat)
        if do_compute_all_dists:
            self.compute_all_dists()
        if logfile is not None:
            logfile.close()
        return learning_error

    def loc_from_idx(self, idx):
        loc = self.locations[idx].view(-1, 2)
        return loc

    def predict(self, dataset, batch_size=100, print_each=100,
                return_density=False, return_errors=False, num_workers=os.cpu_count()):
        """
        Batch the prediction to avoid memory overloading
        """
        dataloader = build_dataloader(dataset, num_workers=num_workers, batch_size=batch_size, shuffle=False)
        bmus = list()
        errors = list()
        if return_density:
            density = np.zeros((self.m, self.n))
        if return_errors:
            bmu_indices = []
        labels = []
        start = time.perf_counter()
        for i, (label, batch) in enumerate(dataloader):
            if not i % print_each:
                print(f'{i}/{len(dataloader)} | time: {str(datetime.timedelta(seconds=time.perf_counter() - start))} ')
            batch = batch.to(self.device)
            labels.extend(label)

            # If we want the topographic error, we need to compute more neighbors, and keep the indices
            bmu_idx, error = self.inference_call(batch, n_bmu=2 if return_errors else 1)
            if return_errors:
                bmu_indices.append(bmu_idx)

            # Then we can keep the first only and compute its bmu affectation
            first_idx = bmu_idx[:, 0] if return_errors else bmu_idx
            bmu_loc = self.loc_from_idx(first_idx)
            if return_density:
                density[tuple(bmu_loc.cpu().numpy().T)] += 1.
            bmus.append(bmu_loc)
            if error.ndim == 0:
                error = error[None, ...]
            errors.append(error)
        bmus = torch.cat(bmus)
        bmus = bmus.cpu().numpy()
        errors = torch.cat(errors)
        errors = errors.cpu().numpy()

        default_return = [bmus, errors, labels]
        if return_density:
            density /= density.sum()
            default_return.append(density)
        # Optionnally compute errors
        if return_errors:
            quantization_error = np.mean(errors[:, :, 0])
            topo_dists = np.array([self.distance_mat[int(first), int(second)] for first, second in bmu_indices])
            topo_error = np.sum(topo_dists > 1) / len(topo_dists)
            print(f'On these samples, the quantization error is {quantization_error:1f} '
                  f'and the topological error rate is {topo_error:1f}')
            default_return.append(quantization_error)
            default_return.append(topo_error)
        return tuple(default_return)

    def plot_component_plane(self, plane, savefig=None, show=True):
        """
        Get component plane plot : The value of each dimension for each centroid
        """
        smap = self.centroids.cpu().numpy().reshape((self.m, self.n, -1))
        assert 0 <= plane < self.dim
        component_plane = smap[:, :, plane]

        plt.matshow(component_plane)
        plt.colorbar()
        if savefig is not None:
            plt.savefig(savefig)
        if show:
            plt.show()
        plt.clf()

    @staticmethod
    def _unfold(umat, adj):
        """
        - umat: U-matrix
        - adj: Adjacency matrix of the corresponding U-matrix
        """
        n1, n2 = umat.shape
        mstree = graph.minimum_spanning_tree(adj)
        start = umat.argmin()
        sdist, pred = graph.shortest_path(mstree, indices=start, directed=False, return_predecessors=True)
        floodpath = np.asarray(np.unravel_index(sdist.argsort(), (n1, n2))).T
        parents_floodpath = np.asarray(np.unravel_index(pred[sdist.argsort()][1:], (n1, n2))).T
        flow = np.sign(floodpath[1:] - parents_floodpath)
        i, j = floodpath[0]
        u, v = 0, 0
        mapping = {(i, j): (u, v)}
        u_min, v_min = 0, 0

        for step, vec in enumerate(flow):
            i, j = floodpath[step + 1]
            u, v = mapping[tuple(parents_floodpath[step])] + vec
            u_min = u if u < u_min else u_min
            v_min = v if v < v_min else v_min
            mapping[(i, j)] = (u, v)
        path = [(u - u_min, v - v_min) for u, v in mapping.values()]
        floodpath = [(i, j) for i, j in floodpath]
        mapping = dict(zip(floodpath, path))
        uumat = np.ones(np.asarray(path).max(axis=0) + 1) * np.inf
        # uumat = np.ones(np.asarray(path).max(axis=0) + 1)
        for k, v in mapping.items():
            uumat[v[0], v[1]] = umat[k[0], k[1]]
        return uumat, mapping

    def _get_umat(self, smap, shape=None, rmsd=False, return_adjacency=False, periodic=True):
        """
        Compute the U-matrix based on a map of centroids and their connectivity.
        """

        def neighbor_dim2_toric(p, s):
            """
            Efficient toric neighborhood function for 2D SOM.
            """
            x, y = p
            X, Y = s
            xm = (x - 1) % X
            ym = (y - 1) % Y
            xp = (x + 1) % X
            yp = (y + 1) % Y
            return [(xm, ym), (xm, y), (xm, yp), (x, ym), (x, yp), (xp, ym), (xp, y), (xp, yp)]

        def neighbor_dim2_grid(p, s):
            """
            Efficient grid neighborhood function for 2D SOM.
            """
            x, y = p
            X, Y = s
            xm = max((x - 1), 0)
            ym = max((y - 1), 0)
            xp = min((x + 1), X - 1)
            yp = min((y + 1), Y - 1)
            shortlist = {(xm, ym), (xm, y), (xm, yp), (x, ym), (x, yp), (xp, ym), (xp, y), (xp, yp)}
            shortlist.discard((x, y))
            return list(shortlist)

        # By default a map is a square
        if shape is None:
            if smap.ndim == 2:
                n = int(np.sqrt(smap.shape[0]))
                smap = smap.reshape((n, n, -1))
            shape = list(smap.shape)[:-1]
        umatrix = np.zeros(shape)
        adjmat = {'data': [], 'row': [], 'col': []}

        for point in itertools.product(*[range(s) for s in shape]):
            neuron = smap[point]
            if periodic:
                neighbors = tuple(np.asarray(neighbor_dim2_toric(point, shape), dtype='int').T)
            else:
                neighbors = tuple(np.asarray(neighbor_dim2_grid(point, shape), dtype='int').T)

            smap_torch, neuron_torch = torch.from_numpy(smap[neighbors]).to(self.device), \
                                       torch.from_numpy(neuron[None]).to(self.device)
            torch_cdists = self.metric(smap_torch, neuron_torch)
            cdists = torch_cdists.cpu().numpy()
            umatrix[point] = cdists.mean()

            adjmat['row'].extend([
                                     np.ravel_multi_index(point, shape),
                                 ] * len(neighbors[0]))
            adjmat['col'].extend(np.ravel_multi_index(neighbors, shape))
            adjmat['data'].extend(cdists[:, 0])
        if rmsd:
            natoms = smap.shape[-1] / 3.
            umatrix /= natoms
            umatrix = np.sqrt(umatrix)

        if return_adjacency:
            adjmat = scipy.sparse.coo_matrix((adjmat['data'], (adjmat['row'], adjmat['col'])),
                                             shape=(np.prod(shape), np.prod(shape)))
            return umatrix, adjmat
        else:
            return umatrix

    def compute_umat(self, unfold=True, normalize=True):
        smap = self.centroids.cpu().numpy().reshape((self.m, self.n, -1))
        umat, adj = self._get_umat(smap, shape=(self.m, self.n), return_adjacency=True, periodic=self.periodic)
        if normalize:
            # Renormalize
            umat = (umat - np.min(umat)) / (np.max(umat) - np.min(umat))
        self.umat = umat
        self.adj = adj
        if self.periodic and unfold:
            uumat, mapping = self._unfold(umat, adj)
            self.uumat = uumat
            self.mapping = mapping
            self.reversed_mapping = {v: k for k, v in self.mapping.items()}
            umat = uumat
        else:
            self.mapping = {(i, j): (i, j) for (i, j) in itertools.product(range(self.m), range(self.n))}
            self.reversed_mapping = {v: k for k, v in self.mapping.items()}
            self.uumat = umat

    def compute_all_dists(self):
        # GRAPH-BASED
        # mstree = graph.minimum_spanning_tree(self.adj)
        adj = self.adj.tocsr()
        self.all_to_all_dist = graph.shortest_path(adj, directed=False)
        # self.all_to_all_dist = graph.shortest_path(mstree, directed=False)

    def cluster(self, min_distance=2):
        """
        Perform clustering based on the umatrix.

        We have tried several methods that can be put into two main categories :
        Graph-based : using either a minimum spanning tree of the full connectivity
        Image-based : using the U-matrix and segmentation techniques. In the periodic case,
                        an unfolding of the umatrix was necessary to do so.
        """

        local_min = peak_local_max(-self.umat, min_distance=min_distance)
        n_local_min = local_min.shape[0]
        clusterer = AgglomerativeClustering(affinity='precomputed', linkage='average', n_clusters=n_local_min)
        try:
            labels = clusterer.fit_predict(self.all_to_all_dist)
        except ValueError as e:
            print(f'WARNING : The following error was catched : "{e}"\n'
                  f'The clusterer yields zero clusters on the data.'
                  ' You should train it more or gather more data')
            labels = np.zeros(self.m * self.n)
        labels = labels.reshape((self.m, self.n))

        # IMAGE-BASED
        # from skimage.morphology import disk
        # from skimage.filters import rank
        # from skimage.util import img_as_ubyte
        # from skimage.segmentation import watershed
        # # denoise image
        # umat = img_as_ubyte(umat)
        # denoised = rank.median(umat, disk(2))
        # # find continuous region (low gradient -
        # # where less than 10 for this image) --> markers
        # # disk(5) is used here to get a more smooth image
        # markers = rank.gradient(denoised, disk(5)) < 10
        # markers = ndi.label(markers)[0]
        # # local gradient (disk(2) is used to keep edges thin)
        # gradient = rank.gradient(denoised, disk(2))
        # labels = watershed(gradient, markers)
        # if self.periodic:
        #     square_label = np.zeros((self.m, self.n))
        #     for (u, v), (i, j) in self.reversed_mapping.items():
        #         square_label[i, j] = labels[u, v]
        #     return square_label
        self.cluster_att = labels.flatten()
        return labels

    def manual_cluster(self, autocluster=False):
        # from .somgui import Wheel, Click

        fig, ax = plt.subplots(figsize=(8, 10))
        cax = ax.matshow(self.umat)
        fig.colorbar(cax)
        plt.show()

        click = Click(ax=ax)
        fig.canvas.mpl_connect('button_press_event', click)
        wheel = Wheel(self, click, ax=ax)

        if autocluster:
            self.clusters_user = self.cluster() + 1
        if self.clusters_user is not None:
            wheel.clusters = self.clusters_user
            wheel.expand_clusters()
            wheel.plot_clusters()
        fig.canvas.mpl_connect('scroll_event', wheel)
        fig.canvas.mpl_connect('button_press_event', wheel)
        ax.format_coord = wheel.format_coord
        plt.show()
        self.cluster_att = wheel.expanded_clusters.flatten()
        self.clusters_user = wheel.clusters

    def predict_cluster(self, samples=None, batch_size=100, user=False):
        """
        we have a mapping from each unit to its cluster in the flattened form in self.cluster_att
        Then we need to turn the bmu attributions into the index in this list and return the cluster attributions
        """
        if self.cluster_att is None:
            cluster_att = self.cluster()
            self.cluster_att = cluster_att.flatten()
        if samples is None:
            if self.bmus is None:
                print('No existing BMUs in the SOM object, one needs data points to predict clusters on')
        else:
            bmus, error, labels = self.predict(samples, batch_size=batch_size)
            self.bmus = bmus
            self.error = error
        flat_bmus = (self.bmus[:, 0] * self.n + self.bmus[:, 1]).astype(np.int32)
        codebook = []
        inds = np.arange(len(self.bmus))
        for i in range(self.m * self.n):
            sel = (flat_bmus == i)
            if sel.sum() > 0:
                ind = inds[sel][self.error[sel].argmin()]
                codebook.append(ind)
            else:
                codebook.append(-1)
        self.codebook = codebook
        if not user:
            return self.cluster_att[flat_bmus], error
        else:
            return self.clusters_user.flatten()[flat_bmus], error

    def get_pairwise_dist(self, num_workers=1, batch_size=10):
        print('Computing pairwise distances between SOM centroids')
        dataloader = build_dataloader(self.centroids.to('cpu'),
                                      num_workers=num_workers,
                                      batch_size=batch_size,
                                      shuffle=False)
        pdist = []
        for i, (label, batch) in enumerate(dataloader):
            sys.stdout.write(f'{i + 1}/{len(dataloader) + 1}\r')
            sys.stdout.flush()
            batch = batch.to(self.device)
            dists = self.metric(batch, self.centroids).flatten()
            pdist.extend(list(dists.to('cpu')))
        pdist = np.asarray(pdist)
        pdist = pdist.reshape((self.m * self.n,) * 2)
        self.pairwise_dist = pdist
        return pdist

    def mds_embedding(self):
        print('MDS embedding')
        if self.pairwise_dist is None:
            self.get_pairwise_dist()
        embedding = MDS(n_components=2, dissimilarity='precomputed', n_jobs=-1)
        pdist = self.pairwise_dist.copy()
        if (pdist < 0).any():
            pdist -= pdist.min()
        if check_symmetric(pdist) != 0.:
            pdist, error = symmetrize(pdist)
            print('Warning: The pairwise distance matrix is not symmetric')
            print(f'Symmetrized array with a relative error of {error:.4g}')
        self.mds = embedding.fit_transform(pdist)
        return self.mds

    def save_pickle(self, outname):
        self.to('cpu')
        pickle.dump(self, open(outname, 'wb'))

    @staticmethod
    def load_pickle(inname, device='cuda' if torch.cuda.is_available() else 'cpu'):
        loaded_som = pickle.load(open(inname, 'rb'))
        loaded_som.to_device(device)
        return loaded_som


def time_som(som, X):
    som.alpha = 0.5
    X = X.to(device)
    import time

    a = time.perf_counter()
    for _ in range(1000):
        som(X[:30], 1)
    if torch.cuda.is_available():
        torch.cuda.synchronize()
    print('total time : ', time.perf_counter() - a)
    sys.exit()


if __name__ == '__main__':
    pass
    # Prepare data
    device = 'cuda' if torch.cuda.is_available() else 'cpu'
    np.random.seed(0)
    X = np.random.rand(1000, 500)
    X = torch.from_numpy(X)
    X = X.float()

    # Create SOM
    n = 10
    somsize = n ** 2
    nsamples = X.shape[0]
    dim = X.shape[1]
    niter = 5
    batch_size = 30
    nsteps = int(nsamples / batch_size)
    som = SOM(n, n, dim, n_epoch=niter, device=device, precompute=True, periodic=True)

    # Fit it and get results
    learning_error = som.fit(X, batch_size=batch_size)
    bmus, inference_error, labels = som.predict(X, batch_size=batch_size)
    predicted_clusts, errors = som.predict_cluster(X[45:56])
    print('some cluster for some random points are : ', predicted_clusts)



#!/usr/bin/env python3
# -*- coding: UTF8 -*-

#############################################################################
# Authors: Vincent Mallet, Guillaume Bouvier	      	                    #
# Copyright (c) 2021 Institut Pasteur                                       #
#                 				                                            #
#                                                                           #
#  Redistribution and use in source and binary forms, with or without       #
#  modification, are permitted provided that the following conditions       #
#  are met:                                                                 #
#                                                                           #
#  1. Redistributions of source code must retain the above copyright        #
#  notice, this list of conditions and the following disclaimer.            #
#  2. Redistributions in binary form must reproduce the above copyright     #
#  notice, this list of conditions and the following disclaimer in the      #
#  documentation and/or other materials provided with the distribution.     #
#  3. Neither the name of the copyright holder nor the names of its         #
#  contributors may be used to endorse or promote products derived from     #
#  this software without specific prior written permission.                 #
#                                                                           #
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS      #
#  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT        #
#  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR    #
#  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT     #
#  HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,   #
#  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT         #
#  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,    #
#  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY    #
#  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT      #
#  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE    #
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.     #
#                                                                           #
#  This program is free software: you can redistribute it and/or modify     #
#                                                                           #
#############################################################################

import os
import torch
import pickle
import matplotlib.pyplot as plt
import numpy as np
from quicksom.som import SOM

from sklearn.datasets import make_moons

try:
    os.mkdir('out')
except FileExistsError:
    pass

try:
    os.mkdir('figs')
except FileExistsError:
    pass

# BUILD DATASET if does not exist yet or too short for the required number of points
max_points = 100
create_data = False
if os.path.exists('data/moons.txt'):
    X = np.genfromtxt('data/moons.txt')
    if len(X) >= max_points:
        create_data = False
        y = X[:, 2]
        X = X[:, :2]

if create_data:
    X, y = make_moons(n_samples=max_points, noise=0.05)
    np.savetxt('data/moons.txt', np.c_[X, y])

X = X[:max_points]
y = y[:max_points]

plt.scatter(X[:, 0], X[:, 1])
plt.savefig('figs/moons.png')

device = 'cuda' if torch.cuda.is_available() else 'cpu'
X = torch.from_numpy(X)
X = X.float()
X = X.to(device)

if not os.path.exists('out/trained.p'):
    m, n = 50, 50
    dim = X.shape[1]
    niter = 100
    batch_size = 100
    som = SOM(m, n, dim, niter=niter, device=device)
    learning_error = som.fit(X, batch_size=batch_size)
    bmus, inference_error = som.predict(X, batch_size=batch_size)
else:
    pass
    som = pickle.load(open('out/trained.p', 'rb'))

som.to_device('cpu')
pickle.dump(som, open('out/trained.p', 'wb'))
X = X.to('cpu')
bmus, inference_error = som.predict(X)
smap = som.centroids.cpu().numpy().reshape((som.m, som.n, -1))
predicted_clusts, errors = som.predict_cluster(X)
umat = som.umat
plt.matshow(umat)
plt.colorbar()
plt.savefig('figs/umat.png')
plt.clf()

# Problem in the x,y scale formalism vs the grid format. First coordinate of an array is the rows...
plt.matshow(umat)
plt.scatter(bmus[:, 1][y == 0], bmus[:, 0][y == 0], c='red')
plt.scatter(bmus[:, 1][y == 1], bmus[:, 0][y == 1], c='green')
plt.savefig('figs/project.png')
plt.clf()

# Problem in the x,y scale formalism vs the grid format. First coordinate of an array is the rows...
plt.matshow(umat)
plt.scatter(bmus[:, 1], bmus[:, 0], c=predicted_clusts)
plt.savefig('figs/project_clusts.png')
plt.clf()

plt.scatter(X[:, 0], X[:, 1], c=predicted_clusts)
plt.savefig('figs/clusts.png')
plt.clf()



#!/usr/bin/env python3
# -*- coding: UTF8 -*-

#############################################################################
# Author: Guillaume Bouvier -- guillaume.bouvier@pasteur.fr                 #
# https://research.pasteur.fr/en/member/guillaume-bouvier/                  #
# Copyright (c) 2022 Institut Pasteur                                       #
#                 				                            #
#                                                                           #
#  Redistribution and use in source and binary forms, with or without       #
#  modification, are permitted provided that the following conditions       #
#  are met:                                                                 #
#                                                                           #
#  1. Redistributions of source code must retain the above copyright        #
#  notice, this list of conditions and the following disclaimer.            #
#  2. Redistributions in binary form must reproduce the above copyright     #
#  notice, this list of conditions and the following disclaimer in the      #
#  documentation and/or other materials provided with the distribution.     #
#  3. Neither the name of the copyright holder nor the names of its         #
#  contributors may be used to endorse or promote products derived from     #
#  this software without specific prior written permission.                 #
#                                                                           #
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS      #
#  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT        #
#  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR    #
#  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT     #
#  HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,   #
#  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT         #
#  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,    #
#  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY    #
#  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT      #
#  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE    #
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.     #
#                                                                           #
#  This program is free software: you can redistribute it and/or modify     #
#                                                                           #
#############################################################################

import streamlit as st
import dill as pickle
import torch
import io
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import re


class CPU_Unpickler(pickle.Unpickler):
    """
    Usage:
    f = open(filename, 'rb')
    contents = CPU_Unpickler(f).load()
    """
    def find_class(self, module, name):
        if module == 'torch.storage' and name == '_load_from_bytes':
            return lambda b: torch.load(io.BytesIO(b), map_location='cpu')
        else:
            return super().find_class(module, name)


def load_som(picklefile):
    som = CPU_Unpickler(picklefile).load()
    return som


def crop_umat(som, vmin=None, vmax=None):
    umat = som.umat.copy()
    if vmin is None:
        vmin = umat.min()
    if vmax is None:
        vmax = umat.max()
    umat = np.clip(a=umat, a_min=vmin, a_max=vmax)
    return umat


def crop_mds(som, vmin, vmax):
    x, y = som.mds.T
    z = crop_umat(som, vmin, vmax).flatten()
    return x, y, z


def stringfilter(string, threshold, arr):
    """
    see: https://stackoverflow.com/a/38986394/1679629
    """
    def matchdist(tosearch, string):
        out = re.search(tosearch, string)
        if out is not None:
            dist = 1. - len(out.group()) / len(tosearch)
        else:
            dist = 1.
        return dist

    dists = np.asarray([matchdist(string, e) for e in arr])
    sel = dists <= threshold
    return sel


def filter_bmus(som, text, threshold=0.):
    labels = som.labels
    sel = stringfilter(string=text, threshold=threshold, arr=labels)
    return sel


if __name__ == '__main__':

    # -- Set page config
    apptitle = 'QuickSOM app'
    st.set_page_config(page_title=apptitle, page_icon=":eyeglasses:")

    # Title the app
    st.title('QuickSOM app')
    uploaded_file = st.file_uploader(label='SOM pickle file to load')
    if uploaded_file is not None:
        som = load_som(uploaded_file)

        # -- Create sidebar for plot controls
        st.sidebar.markdown('## Set Plot Parameters')
        vmin = float(som.umat.min())
        vmax = float(som.umat.max())
        vmin_setter = st.sidebar.slider('vmin', vmin, vmin + (vmax - vmin) / 2., vmin)  # min, max, default
        vmax_setter = st.sidebar.slider('vmax', vmin + (vmax - vmin) / 2., vmax, vmax)  # min, max, default

        # Text input to filter bmus
        instr = st.text_input(label='Filter BMUs based on label')
        bmu_selection = None
        if instr is not None and len(instr) > 0:
            bmu_selection = filter_bmus(som, instr)
            nmatch = bmu_selection.sum()
            st.write(f'{nmatch} data match')

        umat = crop_umat(som, vmin=vmin_setter, vmax=vmax_setter)
        mds = crop_mds(som, vmin=vmin_setter, vmax=vmax_setter)
        fig, axs = plt.subplots(nrows=2, figsize=(10, 10))
        img = axs[0].matshow(umat, cmap='coolwarm', aspect='auto')
        axs[1].scatter(mds[0], mds[1], c=mds[2], cmap='coolwarm')
        if bmu_selection is not None:
            selected_cells = som.bmus[bmu_selection]
            axs[0].scatter(selected_cells[:, 1], selected_cells[:, 0], color='cyan')
            selected_mds = som.mds[np.ravel_multi_index(tuple(selected_cells.T), (som.m, som.n))]
            axs[1].scatter(selected_mds[:, 1], selected_mds[:, 0], color='cyan')
        fig.colorbar(img, ax=axs.ravel().tolist())
        st.pyplot(fig=fig)

        bmus = pd.DataFrame(data={
            'bmu_i': som.bmus[:, 0],
            'bmu_j': som.bmus[:, 1],
            'label': som.labels
        },
                            index=range(len(som.bmus)))



#!/usr/bin/env python
# -*- coding: UTF8 -*-

#############################################################################
# Authors: Vincent Mallet, Guillaume Bouvier	      	                    #
# Copyright (c) 2021 Institut Pasteur                                       #
#                 				                                            #
#                                                                           #
#  Redistribution and use in source and binary forms, with or without       #
#  modification, are permitted provided that the following conditions       #
#  are met:                                                                 #
#                                                                           #
#  1. Redistributions of source code must retain the above copyright        #
#  notice, this list of conditions and the following disclaimer.            #
#  2. Redistributions in binary form must reproduce the above copyright     #
#  notice, this list of conditions and the following disclaimer in the      #
#  documentation and/or other materials provided with the distribution.     #
#  3. Neither the name of the copyright holder nor the names of its         #
#  contributors may be used to endorse or promote products derived from     #
#  this software without specific prior written permission.                 #
#                                                                           #
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS      #
#  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT        #
#  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR    #
#  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT     #
#  HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,   #
#  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT         #
#  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,    #
#  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY    #
#  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT      #
#  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE    #
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.     #
#                                                                           #
#  This program is free software: you can redistribute it and/or modify     #
#                                                                           #
#############################################################################

import argparse
import os
import itertools
import functools
import numpy as np
import time

import torch

import MDAnalysis.coordinates.DCD as md
import pymol.cmd as cmd


def pymol_process(pdb, selection='all', select_align=None):
    """
    Use the PDB file to find the indices in the array for the SOM fitting and the frame alignment.
    :param pdb:
    :param selection:
    :param select_align:
    :return:
    """
    cmd.set('retain_order', 1)
    cmd.load(pdb, object='mystructure')
    pymolspace = {'atoms_ids': [], 'align_ids': []}
    cmd.iterate(selection, 'atoms_ids.append(rank)', space=pymolspace)
    atoms_ids = np.asarray(pymolspace['atoms_ids'])

    # Optionnaly use different indices for alignment.
    if select_align is not None:
        cmd.iterate(select_align, 'align_ids.append(rank)', space=pymolspace)
        align_ids = np.asarray(pymolspace['align_ids'])
    else:
        align_ids = atoms_ids
    return atoms_ids, align_ids


def rigid_body_fit(A, B):
    """
    See: https://en.wikipedia.org/wiki/Kabsch_algorithm
    2-D or 3-D registration with known correspondences.
    Registration occurs in the zero centered coordinate system, and then
    must be transported back.
        Args:
        -    A: Numpy array of shape (N,D) -- Point Cloud to Align (source)
        -    B: Numpy array of shape (N,D) -- Reference Point Cloud (target)
        Returns:
        -    R: optimal rotation
        -    t: optimal translation
    """
    a_mean = A.mean(axis=0)
    b_mean = B.mean(axis=0)
    A_c = A - a_mean
    B_c = B - b_mean
    # Covariance matrix
    H = A_c.T.dot(B_c)
    U, S, Vt = np.linalg.svd(H)
    V = Vt.T
    # Rotation matrix
    R = V.dot(U.T)
    # Translation vector
    t = b_mean - R.dot(a_mean)
    return R, t


def transform(coords, R, t):
    """
    Apply R and a translation t to coords
    """
    coords_out = R.dot(coords.T).T + t
    return coords_out


class DCDataset(torch.utils.data.Dataset):
    def __init__(self, pdbfilename, dcdfilename, selection='polymer', selection_alignment=None):
        """
        This dataset opens a dcd file and make it a PyTorch dataset
        :param dcdfilename:
        """

        self.dcdfilename = dcdfilename
        self.dcdfile = None
        self.fastafile = None
        atoms_ids, align_ids = pymol_process(pdb=pdbfilename, selection=selection, select_align=selection_alignment)
        self.atoms_ids, self.align_ids = atoms_ids, align_ids
        with md.DCDFile(dcdfilename) as dcd:
            self.len = len(dcd)
            first_frame = dcd.read()
            coords = first_frame.xyz
            sel_coords = coords[atoms_ids]
            self.dim = sel_coords.flatten().size
            self.reference_coords = coords[align_ids]

    def __len__(self):
        return self.len

    def __getitem__(self, index):
        """
        Make each worker have a copy of the dcdfile.
        Then for each fetch, use seek to get the right frame, extract the relevant indices.
        Finally use the select_align ones to align to the reference frame and rotate the selected indices accordingly.
        :param index:
        :return:
        """
        if self.dcdfile is None:
            # print(f'opening {self.dcdfilename} with worker id :'
            #       f' {torch.utils.data.get_worker_info().id} and obj id :'
            #       f' {id(self.dcdfilename)} ')
            self.dcdfile = md.DCDFile(self.dcdfilename)

        self.dcdfile.seek(index)
        frame = self.dcdfile.read()
        coords = frame.xyz

        align_coords = coords[self.align_ids]
        R, t = rigid_body_fit(align_coords, self.reference_coords)
        sel_coords = coords[self.atoms_ids]
        aligned_sel_coords = transform(sel_coords, R, t)
        flat_coords = aligned_sel_coords.flatten()
        assert flat_coords.size == self.dim
        return index, flat_coords


def test_parallel(dataset, num_workers, batch_size=10, nloop=100):
    dataloader = torch.utils.data.DataLoader(dataset,
                                             batch_size=batch_size,
                                             shuffle=True,
                                             num_workers=num_workers)
    t0 = time.time()
    dataiter = itertools.cycle(dataloader)
    for i in range(nloop):
        frame_idx, inputvectors = next(dataiter)
    deltat = time.time() - t0
    print(f'Timing for {num_workers} worker(s), batch_size {batch_size} and {nloop} loops: {deltat:.3f} s')
    return deltat


if __name__ == '__main__':
    pass
    pdbname = "data/2lj5.pdb"
    dcdname = "data/2lj5.dcd"
    dataset = DCDataset(dcdfilename=dcdname, pdbfilename=pdbname, selection='name CA')
    test_parallel(num_workers=2, dataset=dataset)



#!/usr/bin/env python3
# -*- coding: UTF8 -*-

#############################################################################
# Authors: Vincent Mallet, Guillaume Bouvier	      	                    #
# Copyright (c) 2021 Institut Pasteur                                       #
#                 				                                            #
#                                                                           #
#  Redistribution and use in source and binary forms, with or without       #
#  modification, are permitted provided that the following conditions       #
#  are met:                                                                 #
#                                                                           #
#  1. Redistributions of source code must retain the above copyright        #
#  notice, this list of conditions and the following disclaimer.            #
#  2. Redistributions in binary form must reproduce the above copyright     #
#  notice, this list of conditions and the following disclaimer in the      #
#  documentation and/or other materials provided with the distribution.     #
#  3. Neither the name of the copyright holder nor the names of its         #
#  contributors may be used to endorse or promote products derived from     #
#  this software without specific prior written permission.                 #
#                                                                           #
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS      #
#  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT        #
#  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR    #
#  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT     #
#  HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,   #
#  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT         #
#  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,    #
#  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY    #
#  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT      #
#  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE    #
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.     #
#                                                                           #
#  This program is free software: you can redistribute it and/or modify     #
#                                                                           #
#############################################################################

import pickle
import argparse
import numpy
import scipy.ndimage
import torch
import matplotlib.pyplot as plt
from skimage.feature import peak_local_max


class Click:
    def __init__(self, ax):
        self.ax = ax
        self.pos = (0, 0)
        self.clickpos, = ax.plot(self.pos[1], self.pos[0], c='r', marker='s')

    def __call__(self, event):
        self.pos = int(event.ydata), int(event.xdata)
        # print(f'i={self.pos[0]}; j={self.pos[1]}')
        self.clickpos.set_data(self.pos[1], self.pos[0])
        self.clickpos.figure.canvas.draw()


def clean_contours(contourplot):
    if contourplot is not None:
        for coll in contourplot.collections:
            coll.remove()


class Wheel:
    def __init__(self, som, click, ax):
        self.ax = ax
        self.threshold = 0
        self.precision = .01
        self.display_str = 'Threshold=%.3f'
        help_string = "\n\n- Left-Click: Set position\n- Scroll: Set basin threshold\n- Shift-Scroll: Set basin threshold with higher precision\n- Left-Click on basin: Set the current basin as a cluster\n- Right-Click on cluster: Delete the selected cluster\n- Double-Click on cluster: Expand the clusters\n- Left-Click on cluster: Come back to user cluster view"
        self.display_str += help_string
        self.threshold_display = self.ax.text(0.5, -0.1, self.display_str % self.threshold,
                                              ha="center", va='top', transform=self.ax.transAxes)
        self.som = som
        self.click = click
        self.pos = self.click.pos
        self.clusters = numpy.zeros((self.som.m, self.som.n), dtype=int)
        self.expanded_clusters = numpy.zeros((self.som.m, self.som.n), dtype=int)
        self.cluster_current = numpy.zeros((self.som.m, self.som.n), dtype=bool)
        self.clusterplot = None  # To plot the current cluster
        self.clustersplot = self.ax.scatter(0, 0, c='r', marker='s', alpha=0.)  # To plot all the clusters

        self.local_min = peak_local_max(-self.som.uumat, min_distance=1)
        self.n_local_min = self.local_min.shape[0]
        self.local_min = numpy.asarray([self.som.reversed_mapping[(e[0], e[1])] for e in self.local_min])
        plt.scatter(self.local_min[:, 1], self.local_min[:, 0], c='g')

    def plot_clusters(self, plot_expanded=False):
        if plot_expanded:
            to_plot = self.expanded_clusters
        else:
            to_plot = self.clusters
        to_plot = numpy.ma.masked_array(data=to_plot, mask=(to_plot == 0))
        self.clustersplot.remove()
        self.clustersplot = self.ax.matshow(to_plot, cmap='tab20')
        self.clustersplot.figure.canvas.draw()

    def __call__(self, event):
        if event.key is 'shift':
            self.precision = 0.001
        else:
            self.precision = 0.01
        if event.button is 'up':
            self.threshold += self.precision
            if self.threshold > 1.:
                self.threshold = 1.
        if event.button is 'down':
            self.threshold -= self.precision
            if self.threshold < 0.:
                self.threshold = 0.
        if event.button == 1:  # Left mouse button pressed
            if self.clusters[self.click.pos] == 0:
                cluster_id = self.clusters.max() + 1
                # print(f"Creating cluster {cluster_id}")
                self.clusters[self.cluster_current] = cluster_id
                self.remap_clusters()
                self.expand_clusters()
            self.plot_clusters()
            if event.dblclick:
                self.plot_clusters(plot_expanded=True)
        if event.button == 3:  # Right mouse button pressed
            cluster_id = self.clusters[self.click.pos]
            if cluster_id > 0:
                self.delete_cluster(cluster_id)
                self.expand_clusters()
                self.plot_clusters()
        self.threshold_display.set_text(self.display_str % self.threshold)
        self.cluster()
        plt.draw()

    @property
    def is_cluster(self):
        return self.cluster_current[self.click.pos]

    def delete_cluster(self, cluster_id):
        # print(f"Deleting cluster {cluster_id}")
        self.clusters[self.clusters == cluster_id] = 0
        self.remap_clusters()

    @property
    def cluster_ids(self):
        cluster_ids = numpy.unique(self.clusters)
        cluster_ids = cluster_ids[cluster_ids > 0]
        return cluster_ids

    def remap_clusters(self):
        remap = numpy.zeros_like(self.clusters)
        for i, cid in enumerate(self.cluster_ids):
            sel = (self.clusters == cid)
            remap[sel] = i + 1
        self.clusters = remap
        print(f'Clusters: {numpy.unique(self.cluster_ids)}')

    def cluster(self):
        pos = self.click.pos
        # ind = numpy.ravel_multi_index(pos, (self.som.m, self.som.n))
        # dists = self.som.all_to_all_dist[ind][numpy.ravel_multi_index(self.local_min.T, (self.som.m, self.som.n))]
        # pos = tuple(self.local_min[numpy.argmin(dists)])
        if self.pos != pos:
            self.threshold = self.som.umat[pos]
        uclusters = self.som.uumat < self.threshold
        label, num_features = scipy.ndimage.label(uclusters)
        label_id = label[self.som.mapping[pos]]
        clusters = numpy.zeros_like(self.som.umat, dtype=bool)
        if label_id > 0:
            zone = (label == label_id)
            uclusters = zone
            for i in range(self.som.m):
                for j in range(self.som.n):
                    u, v = self.som.mapping[(i, j)]
                    clusters[i, j] = uclusters[u, v]
        clean_contours(self.clusterplot)
        self.clusterplot = self.ax.contour(clusters, levels=1, colors='w')
        self.cluster_current = numpy.copy(clusters)
        self.pos = pos

    def expand_clusters(self):
        n_clusters = len(self.cluster_ids)
        expanded_clusters = numpy.copy((self.clusters)).flatten()
        if n_clusters > 1:
            voidcells = numpy.where((self.clusters == 0).flatten())[0]
            for cell in voidcells:
                neighbors = self.som.all_to_all_dist[cell].argsort()
                neighbor_clusters = self.clusters.flatten()[neighbors]
                cluster_assigned = neighbor_clusters[neighbor_clusters > 0][0]
                expanded_clusters[cell] = cluster_assigned
        self.expanded_clusters = expanded_clusters.reshape((self.som.m, self.som.n))

    def format_coord(self, x, y):
        return f'i={int(y)}, j={int(x)}, cluster {self.expanded_clusters[int(y), int(x)]}'


if __name__ == '__main__':
    from quicksom.som import SOM
    parser = argparse.ArgumentParser(description='SOM Graphical User Interface')
    parser.add_argument('--som', type=str, help='SOM pickle object file',
                        required=True)
    args = parser.parse_args()
    sompickle = args.som
    device = 'cuda' if torch.cuda.is_available() else 'cpu'
    som = pickle.load(open(sompickle, 'rb'))
    som.to_device(device)
    # som.cluster()

    fig, ax = plt.subplots(figsize=(8, 10))
    cax = ax.matshow(som.umat)
    fig.colorbar(cax)

    click = Click(ax=ax)
    fig.canvas.mpl_connect('button_press_event', click)
    wheel = Wheel(som, click, ax=ax)
    if som.clusters_user is not None:
        wheel.clusters = som.clusters_user
        wheel.expand_clusters()
        wheel.plot_clusters()
    fig.canvas.mpl_connect('scroll_event', wheel)
    fig.canvas.mpl_connect('button_press_event', wheel)
    ax.format_coord = wheel.format_coord
    plt.show()
    som.cluster_att = wheel.expanded_clusters.flatten()
    som.clusters_user = wheel.clusters
    som.to_device('cpu')
    pickle.dump(som, open(sompickle, 'wb'))



import setuptools

with open("quicksom/README.md", "r") as fh:
    long_description = fh.read()

requirements = ["numpy>1.0",
                "scipy>0.1",
                "scikit_image>0.15",
                "scikit_learn>0.22"]


setuptools.setup(
    name="quicksom",
    version="1.0.0",
    author="Vincent Mallet and Michael Nilges and Guillaume Bouvier",
    author_email="vincent.mallet96@gmail.com",
    description="Self Organizing Maps efficient implementation using PyTorch and JAX",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/bougui505/quicksom",
    packages=setuptools.find_packages(),
    install_requires=requirements,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.5',
    scripts=['quicksom/bin/quicksom_fit',
             'quicksom/bin/quicksom_gui',
             'quicksom/bin/quicksom_predict',
             'quicksom/bin/quicksom_extract',
             'quicksom/bin/quicksom_project',
             'quicksom/bin/quicksom_flow',
             'quicksom/bin/mdx',
             'quicksom/bin/dcd2npy']
)



#!/usr/bin/env python3

import pickle
import sys

import numpy
import torch

sys.path.append('../')
from som import SOM

import argparse

parser = argparse.ArgumentParser()
# In/Out
parser.add_argument("-o", "--out_name", default='data/som.p', help="name of pickle to dump")
args, _ = parser.parse_known_args()

device = 'cuda' if torch.cuda.is_available() else 'cpu'

X = numpy.load('data/x.npy')
X = torch.from_numpy(X)
X = X.float()
X = X.to(device)
dim = X.shape[1]
print(X.shape)
# som = pickle.load(open(args.out_name, 'rb'))
som = SOM(10, 5, dim, niter=10, device=device)
som.fit(X)
quantization_error, topo_error = som.compute_error(samples=X)
som.plot_component_plane(plane=1)
